from __future__ import annotations

from relationalai.early_access.metamodel import ir, factory as f, helpers
from relationalai.early_access.metamodel.visitor import Rewriter
from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set, NameCache
from relationalai.early_access.metamodel import dependency

class ExtractNestedLogicals(Pass):

    def __init__(self):
        super().__init__()

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # process the root node
        extractor = LogicalExtractor(model)
        root = extractor.walk(model.root, model)

        # no extractions, just return the original model
        if not extractor.ctx.top_level:
            return model

        # create a new model with the extracted top level + the new root
        body = extractor.ctx.top_level + [root]
        return ir.Model(
                model.engines,
                OrderedSet.from_iterable(model.relations).update(extractor.ctx.relations).frozen(),
                model.types,
                ir.Logical(model.root.engine, tuple(), tuple(body))
            )

class LogicalExtractor(Rewriter):
    def __init__(self, model):
        super().__init__()
        self.ctx = helpers.RewriteContext()
        self.info = dependency.analyze(model.root)
        self.name_cache = NameCache()

    def handle_logical(self, node: ir.Logical, parent: ir.Node):
        # rewrite the children
        logical = super().handle_logical(node, parent)

        # logicals that hoist vars and all vars do not have a default value will be
        # extracted, except for logicals that require special treatment of the exposed
        # variables (which is currently done by flatten), such as when the parent is a Match
        # or a Union, of if the logical has a Rank.
        if not (
            node.hoisted and
            not isinstance(parent, (ir.Match, ir.Union)) and
            all(isinstance(v, ir.Var) for v in node.hoisted) and
            not any(isinstance(c, ir.Rank) for c in node.body)
            ):
            return logical

        # compute the vars to be exposed by the extracted logical; those are keys (what
        # makes the values unique) + the values (the hoisted variables)
        exposed_vars = ordered_set()
        # start with all the inputs to this logical as keys
        # TODO - in the future we can analyze better these inputs to see if we can drop
        # some and have narrower intermediate relations
        exposed_vars.update(self.info.task_inputs(logical))
        # if there are aggregations, make sure we don't expose the projected and input vars,
        # but expose groupbys
        for child in node.body:
            if isinstance(child, ir.Aggregate):
                exposed_vars.difference_update(child.projection)
                exposed_vars.difference_update(helpers.aggregate_inputs(child))
                exposed_vars.update(child.group)
        # add the values (hoisted)
        exposed_vars.update(helpers.hoisted_vars(logical.hoisted))

        body = ordered_set()
        body.update(self.info.task_dependencies(node)) # notice info is based on the original node
        body.update(logical.body)

        name = helpers.create_task_name(self.name_cache, logical, "_nested_logical")
        connection = helpers.extract(logical, body, exposed_vars.get_list(), self.ctx, name)
        return f.lookup(connection, exposed_vars.get_list())



from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.utils import UniqueNames
from relationalai.early_access.lqp.constructors import mk_primitive, mk_type, mk_var

rel_to_lqp = {
    "=": "rel_primitive_eq",
    "!=": "rel_primitive_neq",
    "%": "rel_primitive_remainder",
    "abs": "rel_primitive_abs",
    "ceil": "rel_primitive_round_up",
    "floor": "rel_primitive_round_down",
    "construct_date": "rel_primitive_construct_date",
    "construct_date_from_datetime": "rel_primitive_datetime_date_convert",
    "construct_datetime": "rel_primitive_construct_datetime",
    "construct_datetime_ms_tz": "rel_primitive_construct_datetime",
    "hash": "rel_primitive_hash_tuple_uint128",
    "uuid_to_string": "rel_primitive_uuid_string",
    "parse_date": "rel_primitive_parse_date",
    "parse_datetime": "rel_primitive_parse_datetime",
    "parse_decimal": "rel_primitive_parse_decimal",
    "parse_int64": "rel_primitive_parse_int",
    "parse_int128": "rel_primitive_parse_int128",
    "string": "rel_primitive_string",
    "starts_with": "rel_primitive_starts_with",
    "ends_with": "rel_primitive_ends_with",
    "contains": "rel_primitive_contains",
    "num_chars": "rel_primitive_num_chars",
    "substring": "rel_primitive_substring",
    "like_match": "rel_primitive_like_match",
    "lower": "rel_primitive_lowercase",
    "upper": "rel_primitive_uppercase",
    "concat": "rel_primitive_concat",
    "replace": "rel_primitive_replace",
    "strip": "rel_primitive_trim",
    "date_year": "rel_primitive_date_year",
    "date_month": "rel_primitive_date_month",
    "date_day": "rel_primitive_date_day",
    "date_add": "rel_primitive_typed_add_date_period",
    "date_subtract": "rel_primitive_typed_subtract_date_period",
    "datetime_add": "rel_primitive_typed_add_datetime_period",
    "datetime_subtract": "rel_primitive_typed_subtract_datetime_period",
    "dates_period_days": "rel_primitive_date_days_between",
    "datetimes_period_milliseconds": "rel_primitive_datetime_milliseconds_between",
    "date_format": "rel_primitive_format_date",
    "datetime_format": "rel_primitive_format_datetime",
    "range": "rel_primitive_range",
    "natural_log": "rel_primitive_natural_log",
    "sqrt": "rel_primitive_sqrt",
    "isinf": "rel_primitive_isinf",
    "isnan": "rel_primitive_isnan",
    # Division is monotype, but only on the input args. Until we distinguish between input
    # and output args, we can't use the same assertions for monotype-ness as the other ops.
    "/": "rel_primitive_divide_monotype",
}

# Mappings of primitive names to their lqp types that represent the raicode expectations.
primitive_types = {
    "date_year": [lqp.TypeName.DATE, lqp.TypeName.INT],
    "date_month": [lqp.TypeName.DATE, lqp.TypeName.INT],
    "date_day": [lqp.TypeName.DATE, lqp.TypeName.INT],
}

agg_to_lqp = {
    "min": "rel_primitive_min",
    "max": "rel_primitive_max",
    "sum": "rel_primitive_add_monotype",
    "count": "rel_primitive_add_monotype", # count is a sum of 1s
    "rel_primitive_solverlib_ho_appl": "rel_primitive_solverlib_ho_appl",
}

rel_to_lqp_monotype = {
    "+": "rel_primitive_add_monotype",
    "-": "rel_primitive_subtract_monotype",
    "*": "rel_primitive_multiply_monotype",
    "<=": "rel_primitive_lt_eq_monotype",
    ">=": "rel_primitive_gt_eq_monotype",
    ">": "rel_primitive_gt_monotype",
    "<": "rel_primitive_lt_monotype",
    "//": "rel_primitive_trunc_divide_monotype",
    "maximum": "rel_primitive_max",
    "minimum": "rel_primitive_min",
}

def relname_to_lqp_name(name: str) -> str:
    # TODO: do these proprly
    if name in rel_to_lqp:
        return rel_to_lqp[name]
    elif name in rel_to_lqp_monotype:
        return rel_to_lqp_monotype[name]
    else:
        # If we don't have a mapping for the built-in, we just pass it through as-is.
        return name

def is_monotype(name: str) -> bool:
    return name in rel_to_lqp_monotype

# We take the name and type of the variable that we're summing over, so that we can generate
# recognizable names for the variables in the reduce operation and preserve the type.
def lqp_avg_op(names: UniqueNames, op_name: str, sum_name: str, sum_type: lqp.Type) -> lqp.Abstraction:
    count_type = mk_type(lqp.TypeName.INT)
    vars = [
        (mk_var(names.get_name(sum_name)), sum_type),
        (mk_var(names.get_name("counter")), count_type),
        (mk_var(names.get_name(sum_name)), sum_type),
        (mk_var(names.get_name("one")), count_type),
        (mk_var(names.get_name("sum")), sum_type),
        (mk_var(names.get_name("count")), count_type),
    ]

    x1 = vars[0][0]
    x2 = vars[1][0]
    y1 = vars[2][0]
    y2 = vars[3][0]
    sumv = vars[4][0]
    count = vars[5][0]

    body = lqp.Conjunction(
        args=[
            mk_primitive("rel_primitive_add_monotype", [x1, y1, sumv]),
            mk_primitive("rel_primitive_add_monotype", [x2, y2, count])
        ],
        meta=None
    )
    return lqp.Abstraction(vars=vars, value=body, meta=None)

# Default handler for aggregation operations in LQP.
def lqp_agg_op(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
    x = mk_var(names.get_name(f"x_{aggr_arg_name}"))
    y = mk_var(names.get_name(f"y_{aggr_arg_name}"))
    z = mk_var(names.get_name(f"z_{aggr_arg_name}"))
    ts = [(x, aggr_arg_type), (y, aggr_arg_type), (z, aggr_arg_type)]

    name = agg_to_lqp.get(op_name, op_name)
    body = mk_primitive(name, [x, y, z])

    return lqp.Abstraction(vars=ts, value=body, meta=None)

def lqp_operator(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
    # TODO: Can we just pass through unknown operations?
    if op_name not in agg_to_lqp:
        raise NotImplementedError(f"Unsupported aggregation: {op_name}")

    return lqp_agg_op(names, op_name, aggr_arg_name, aggr_arg_type)



from __future__ import annotations
from typing import Any, Union
import textwrap

from relationalai.early_access.builder import builder as b, std
from relationalai.early_access.rel.executor import RelExecutor
from .common import make_name
from relationalai.experimental.solvers import Solver

_Any = Union[b.Producer, str, float, int]
_String = Union[b.Producer, str]
_Number = Union[b.Producer, float, int]

class SolverModelDev:
    def __init__(self, model: b.Model, num_type: str):
        self._model = model
        assert num_type in ("cont", "int")
        data_type = "int" if num_type == "int" else "float"
        self._data_type = data_type
        self._variable_relationships = set()
        self._expr_id = 0
        Relationship = model.Relationship

        # TODO workaround for QB not distinguishing string names for Concepts/Relationships, only a problem with raw Rel for result extraction maybe
        sm_id = next(b._global_id)
        def _name(s: str) -> str:
            return f"{s}_{sm_id}"

        # variable relations
        self.variables = Relationship("{node} is variable", short_name=_name("variables"))
        self.variable_type = Relationship("{node} has {type:str}", short_name=_name("variable_type")) # TODO use codes/enum
        self.variable_name = Relationship("{node} has {name:str}", short_name=_name("variable_name"))
        self.variable_start = Relationship(f"{{node}} has {{start:{data_type}}}", short_name=_name("variable_start"))

        # expression relations
        self.operator = Relationship("{node} has {op:str}", short_name=_name("operator")) # TODO use codes/enum
        self.ordered_args_hash = Relationship("{node} arg {i:int} is {arg}", short_name=_name("ordered_args_hash"))
        self.ordered_args_data = Relationship(f"{{node}} arg {{i:int}} is {{arg:{data_type}}}", short_name=_name("ordered_args_data"))
        self.unordered_args_hash = Relationship("{node} arg {i} is {arg}", short_name=_name("unordered_args_hash"))
        self.min_objectives = Relationship("{node} is min objective", short_name=_name("min_objectives"))
        self.max_objectives = Relationship("{node} is max objective", short_name=_name("max_objectives"))
        self.constraints = Relationship("{node} is constraint", short_name=_name("constraints"))
        self.expression_name = Relationship("{node} has {name:str}", short_name=_name("expression_name")) # TODO use codes/enum

        # solver result relations
        self.result_info = model.Relationship("{key:str} has {val:str}", short_name=_name("result_info"))
        self.point = model.Relationship(f"{{var}} has {{point:{data_type}}}", short_name=_name("point"))
        self.points = model.Relationship(f"point {{i:int}} for {{var}} has {{point:{data_type}}}", short_name=_name("points"))

    def solve_for(self, expr: b.Relationship | b.Fragment, populate: bool = True, **kwargs):
        where = []
        if isinstance(expr, b.Fragment):
            assert expr._select and len(expr._select) == 1 and expr._where, "Fragment input for `solve_for` must have exactly one select and a where clause"
            rel = expr._select[0]
            where.extend(expr._where)
        elif isinstance(expr, b.Relationship):
            rel = expr
        else:
            raise ValueError(f"Invalid expression type {type(expr)} for `solve_for`; must be a Relationship or Fragment")
        assert rel._parent and rel._short_name, "Relationship for `solve_for` must have a parent and a short name"
        assert rel not in self._variable_relationships

        self._variable_relationships.add(rel)
        node = b.Hash.ref()
        defs = [self.variables(node)]

        # handle optional variable properties
        new_kwargs = kwargs.copy()
        if "type" not in new_kwargs:
            new_kwargs["type"] = "cont" if self._data_type == "float" else "int"
        for (key, val) in new_kwargs.items():
            if key == "name":
                assert isinstance(val, (_Any, list)), f"Expected {key} to be a value or list, got {type(val)}"
                defs.append(self.variable_name(node, make_name(val)))
            elif key == "type":
                assert val in ("cont", "int", "bin"), f"Unsupported variable type {val} for `solve_for`; must be cont, int, or bin"
                defs.append(self.variable_type(node, val))
            elif key in ("lower", "upper", "fixed", "start"):
                assert isinstance(val, _Number), f"Expected {key} to be a number, got {type(val)}"
                if key == "start":
                    defs.append(self.variable_start(node, val))
                else:
                    # TODO use lower/upper/fixed tables in future rather than making constraints
                    op = ">=" if key == "lower" else ("<=" if key == "upper" else "=")
                    self.satisfy(b.require(b.Expression(b.Relationship.builtins[op], rel, val)).where(*where))
            else:
                raise ValueError(f"Invalid keyword argument {key} for `solve_for`")

        where.append(_make_hash((rel._short_name, rel._parent), node))
        b.define(*defs).where(*where)

        if populate:
            # get variable values from the result point (populated by the solver)
            # TODO maybe instead delete/insert into variable relationships after solve.
            val = b.Number.ref()
            b.define(rel(val)).where(self.point(node, val), *where)

        return None

    def minimize(self, expr: _Number, name: _String | list | None = None):
        assert isinstance(expr, _Number)
        return self._handle_expr(self.min_objectives, expr, name)

    def maximize(self, expr: _Number, name: _String | list | None = None):
        assert isinstance(expr, _Number)
        return self._handle_expr(self.max_objectives, expr, name)

    def satisfy(self, expr: b.Fragment, check: bool = False, name: _String | list | None = None):
        assert expr._require, "Fragment input for `satisfy` must have a require clause"
        assert not expr._select and not expr._define, "Fragment input for `satisfy` must not have a select or define clause"
        if not check:
            # remove the `require` from the model roots so it is not checked
            b._remove_roots([expr])
        for req in expr._require:
            self._handle_expr(self.constraints, req, name, where=expr._where)
        return None

    def _handle_expr(self, root_type, expr, name: _String | list | None, where: list = []):
        ctx = ExprContext(self)
        ctx.where.extend(where)
        root = _rewrite(expr, ctx)
        assert root is not None, f"Cannot symbolify {expr}"
        ctx.define.append(root_type(root))
        # if name is not None:
        #     # TODO not printing expression names yet, so skip for now
        #     assert isinstance(name, (_String, list)), f"Expected name to be a string or list, got {type(name)}"
        #     ctx.define.append(self.expression_name(root, make_name(name)))
        self._handle_ctx(ctx)
        return None

    def _handle_ctx(self, ctx: ExprContext, where: list = []):
        all_where = ctx.where + where
        b.define(*ctx.define).where(*all_where)
        for subctx in ctx.subcontext:
            self._handle_ctx(subctx, where=all_where)

    # print the variables and components of the model in human-readable format
    def print(self, verbose: bool = True, expr_names: bool = False):
        # summarize variables
        var_types = ["int", "bin", "cont"]
        vi = b.Hash.ref()
        var_counts = b.select(*[b.count(vi).where(self.variable_type(vi, t)) | 0 for t in var_types]).to_df()
        assert var_counts.shape == (1, 3)
        (int_count, bin_count, cont_count) = var_counts.iloc[0]

        # summarize expressions
        expr_types = [
            (self.min_objectives, "Minimization objectives"),
            (self.max_objectives, "Maximization objectives"),
            (self.constraints, "Constraints"),
        ]
        ei = b.Hash.ref()
        expr_counts = b.select(*[b.count(ei).where(E(ei)) | 0 for (E, _) in expr_types]).to_df()
        assert expr_counts.shape == (1, 3)
        (min_count, max_count, cons_count) = expr_counts.iloc[0]

        # print summary
        print("Solver model has:")
        print(f"• {int_count} integer variables, {bin_count} binary variables, {cont_count} continuous variables")
        print(f"• {min_count} minimization objectives, {max_count} maximization objectives, {cons_count} constraints")
        if not verbose:
            return None

        # print variable names
        vn, vt = b.String.ref(), b.String.ref()
        var_names = b.select(vn.alias("name"), vt.alias("type")).where(self.variable_name(vi, vn), self.variable_type(vi, vt)).to_df()
        assert not var_names.empty, "No variable names found in the model"
        print("Variables:")
        print(var_names.to_string(index=False, header=True))

        # print expressions TODO print expression names optionally
        for (E, s) in expr_types:
            expr_strings = self._expr_strings(E)
            if expr_strings:
                print(s + ":")
                print("\n".join(expr_strings))
        return None

    def _expr_strings(self, roots):
        # materialize data
        iv = b.Hash.ref()
        roots_df = b.select(iv).where(roots(iv)).to_df()
        if roots_df.empty:
            return []
        nodes = roots_df.iloc[:, 0].tolist()

        sv = b.String.ref()
        names_dict = {k: v for k, v in b.select(iv, sv).where(self.variable_name(iv, sv))}
        ops_dict = {k: v for k, v in b.select(iv, sv).where(self.operator(iv, sv))}

        args_dict = {}
        iv2 = b.Hash.ref()
        iv3 = b.Hash.ref()
        # unordered args
        for k, _, v in b.select(iv, iv2, iv3).where(self.unordered_args_hash(iv, iv2, iv3)):
            if k not in args_dict:
                args_dict[k] = []
            args_dict[k].append(v)
        ordered_args_dict = {}
        iv2 = b.Integer.ref()
        for k, i, v in b.select(iv, iv2, iv3).where(self.ordered_args_hash(iv, iv2, iv3)):
            if k not in ordered_args_dict:
                ordered_args_dict[k] = []
            ordered_args_dict[k].append((i, v))
        vv = b.python_types_str_to_concepts[self._data_type].ref()
        for k, i, v in b.select(iv, iv2, vv).where(self.ordered_args_data(iv, iv2, vv)):
            if k not in ordered_args_dict:
                ordered_args_dict[k] = []
            ordered_args_dict[k].append((i, v))
        # convert ordered args to dict
        for k, t in ordered_args_dict.items():
            # sort by index
            t.sort(key=lambda x: x[0])
            args_dict[k] = [v for (_, v) in t]

        # compute strings
        return sorted(_expr_strings_rec(n, names_dict, ops_dict, args_dict) for n in nodes)

    # solve the model given a solver and solver options
    # TODO for now we are materializing the expression graph as dataframes and writing to CSVs in python on the client side, then sending the CSVs along with the solver options as HTTP parameters, skipping the export of a model file from the engine.
    def solve(self, solver: Solver, log_to_console: bool = False, **kwargs):
        # validate options
        for k, v in kwargs.items():
            if not isinstance(k, str):
                raise ValueError(f"Invalid parameter key. Expected string, got {type(k)} for {k}.")
            if not isinstance(v, (int, float, str, bool)):
                raise ValueError(
                    f"Invalid parameter value. Expected string, integer, float, or boolean, got {type(v)} for {k}."
                )

        # 1. Query the model relations and convert to CSV strings.
        print("query model relations")
        def _csv(t):
            return t.to_df().to_csv(index=False, header=False)
        model_relations: dict[str, Any] = {
            "variable_type": _csv(self.variable_type),
            "variable_name": _csv(self.variable_name),
            "operator": _csv(self.operator),
            "ordered_args_hash": _csv(self.ordered_args_hash),
            "ordered_args_data": _csv(self.ordered_args_data),
            "unordered_args_hash": _csv(self.unordered_args_hash),
            "min_objectives": _csv(self.min_objectives),
            "max_objectives": _csv(self.max_objectives),
            "constraints": _csv(self.constraints),
        }

        # 2. Execute job and wait for completion.
        print("execute solver job")
        kwargs["model_relations"] = model_relations
        payload = {"solver": solver.solver_name.lower(), "options": kwargs}
        job_id = solver._exec_job(payload, log_to_console=log_to_console)

        # 3. Extract result.
        print("extract result")
        executor = self._model._to_executor()
        assert isinstance(executor, RelExecutor)
        executor.execute_raw(textwrap.dedent(f"""
        def raw_result {{
            load_binary["snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-results/{job_id}/result.binpb"]
        }}

        ic result_not_empty("Solver result is empty.") requires not empty(raw_result)

        @no_diagnostics(:EXPERIMENTAL)
        def extracted {{ rel_primitive_solverlib_extract[raw_result] }}

        def delete[:{self.result_info._name}]: {self.result_info._name}
        def delete[:{self.point._name}]: {self.point._name}
        def delete[:{self.points._name}]: {self.points._name}

        def insert(:{self.result_info._name}, key, val):
            exists((k) | string(extracted[k], val) and ::std::mirror::lower(k, key))
        def insert(:{self.point._name}, var, val): extracted(:point, var, val)
        def insert(:{self.points._name}, i, var, val): extracted(:points, var, i, val)
        """), readonly=False)

        print("finished solve")
        return None

    # load a particular point index from `points` into `point`
    # so it is accessible from the variable relationship
    def load_point(self, i: int):
        if not isinstance(i, int) and i >= 0:
            raise ValueError(f"Expected nonnegative integer index for point, got {i}")
        executor = self._model._to_executor()
        assert isinstance(executor, RelExecutor)
        executor.execute_raw(textwrap.dedent(f"""
        def delete[:{self.point._name}]: {self.point._name}
        def insert(:{self.point._name}, var, val): {self.points._name}(int128[{i}], var, val)
        """), readonly=False)
        return None

    # print summary of the solver result
    def summarize_result(self):
        to_get = ["error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "result_count"]
        k, v = b.String.ref(), b.String.ref()
        df = b.select(k, v).where(self.result_info(k, v), k.in_(to_get)).to_df()
        print(df.to_string(index=False, header=False))
        return df

    # select variable names and values in the primal result point(s)
    def variable_values(self, multiple: bool = False):
        var = b.Hash.ref()
        val = b.Number.ref()
        s = b.String.ref()
        if multiple:
            i = b.Integer.ref()
            return b.select(i, s, val).where(self.points(i, var, val), self.variable_name(var, s))
        else:
            return b.select(s, val).where(self.point(var, val), self.variable_name(var, s))

    # get scalar information
    def __getattr__(self, name: str):
        df = None
        # model info
        if name in {"num_variables", "num_constraints", "num_min_objectives", "num_max_objectives"}:
            map = {
                "num_variables": self.variables,
                "num_constraints": self.constraints,
                "num_min_objectives": self.min_objectives,
                "num_max_objectives": self.max_objectives,
            }
            node = b.Hash.ref()
            df = b.select(b.count(node).where(map[name](node)) | 0).to_df()
        # result info
        if name in {"error", "termination_status", "solver_version", "printed_model", "solve_time_sec", "objective_value", "result_count"}:
            val = b.String.ref()
            df = b.select(val).where(self.result_info(name, val)).to_df()
        if df is not None:
            if not df.shape == (1, 1):
                raise ValueError(f"Expected exactly one value for {name}, but df has shape {df.shape}")
            v = df.iloc[0, 0]
            if isinstance(v, str):
                if name == "solve_time_sec":
                    return float(v)
                elif name == "objective_value":
                    return int(v) if self._data_type == "int" else float(v)
                elif name == "result_count":
                    return int(v)
            return v
        return None


class ExprContext():
    def __init__(self, solver_model):
        self.solver_model = solver_model
        self.define = []
        self.where = []
        self.subcontext = []

def _rewrite(expr: b.Producer | b.Fragment, ctx: ExprContext):
    if isinstance(expr, (int, float, str)):
        return None

    elif isinstance(expr, b.Concept):
        return None

    elif isinstance(expr, b.Ref):
        thing = _rewrite(expr._thing, ctx)
        if thing:
            assert isinstance(thing, b.Producer)
            return thing.ref()
        return None

    elif isinstance(expr, (b.Relationship, b.RelationshipRef, b.RelationshipFieldRef)):
        rel = expr if isinstance(expr, b.Relationship) else expr._relationship
        if rel in ctx.solver_model._variable_relationships:
            return std.hash(rel._short_name, expr._parent)
        return None

    elif isinstance(expr, b.Union):
        # return union of the symbolified expressions, if any are symbolic
        args_rewritten = False
        args = []
        for arg in expr._args:
            ra = _rewrite(arg, ctx)
            if ra:
                args_rewritten = True
                args.append(ra)
            else:
                args.append(arg)
        if args_rewritten:
            return b.union(*args)
        return None

    elif isinstance(expr, b.Expression):
        assert isinstance(expr._op, b.Relationship)
        assert _rewrite(expr._op, ctx) is None # shouldn't contain variables
        sym_args = [_rewrite(a, ctx) for a in expr._params]
        if all(a is None for a in sym_args):
            return None
        op = expr._op._name
        assert op in infixs or op in prefix_ops, f"Operator {op} not supported in first order solver rewrites"

        # some arguments involve solver variables, so rewrite the expression
        new_args = sym_args if op in infix_comps else sym_args[:-1] # drop last arg introduced by builder
        sm = ctx.solver_model
        node = b.Hash.ref()
        sm._expr_id += 1
        ctx.define.append(sm.operator(node, op))
        hash_args = []
        for (i, a) in enumerate(new_args):
            if a is None:
                d = sm.ordered_args_data(node, i, expr._params[i])
                hash_args.append(expr._params[i])
            else:
                d = sm.ordered_args_hash(node, i, a)
                hash_args.append(a)
            ctx.define.append(d)
        ctx.where.append(_make_hash((sm._expr_id, *hash_args), node))
        return node

    elif isinstance(expr, b.Aggregate):
        # only the last argument can be symbolic
        pre_args = expr._args[:-1]
        sym_arg = _rewrite(expr._args[-1], ctx)
        if not sym_arg:
            return None
        op = expr._op._name
        assert op in agg_ops, f"Operator {op} not supported in aggregate rewrites"

        sm = ctx.solver_model
        node = b.Hash.ref()
        ctx.where.append(_make_hash((sm._expr_id, *expr._group), node))
        # only apply where conditions in subcontext
        subctx = ExprContext(sm)
        ctx.subcontext.append(subctx)
        subctx.where.extend(expr._where._where)
        arg_hash = b.Hash.ref()
        subctx.where.append(_make_hash((sm._expr_id, *pre_args), arg_hash)) # TODO also add sym_arg here?
        sm._expr_id += 1
        subctx.define.append(sm.operator(node, op))
        subctx.define.append(sm.unordered_args_hash(node, arg_hash, sym_arg)) # TODO what if some values are data not hashes?
        return node

    elif isinstance(expr, b.Fragment):
        # only support selects with one item
        assert not expr._define and not expr._require and len(expr._select) == 1
        sym_select = _rewrite(expr._select[0], ctx)
        if sym_select:
            return b.select(sym_select).where(*expr._where)
        return None

    raise NotImplementedError(f"Solver rewrites cannot handle {expr} of type {type(expr)}")

def _expr_strings_rec(x, names_dict, ops_dict, args_dict):
    if x in names_dict:
        # this is a variable node
        return names_dict[x]
    elif x not in ops_dict:
        # this is a value
        return str(x)

    # this is an expression
    assert x in ops_dict
    arg_strs = []
    op = ops_dict[x]

    for k in args_dict.get(x, []):
        s = _expr_strings_rec(k, names_dict, ops_dict, args_dict)
        if op in infix_ops and k in ops_dict and ops_dict[k] in infixs:
            # add parentheses to avoid precedence issues
            s = f"({s})"
        arg_strs.append(s)

    if op in agg_ops:
        # sort unordered args to improve determinism
        arg_strs.sort()

    if op in infixs:
        # infix operator, e.g. x + y
        assert len(arg_strs) == 2 # TODO could relax e.g. for +, *
        return f"{arg_strs[0]} {op} {arg_strs[1]}"
    else:
        # prefix operator, e.g. abs(x)
        assert len(arg_strs) >= 1
        return f"{op}({', '.join(arg_strs)})"

def _make_hash(tup, res):
    return b.Expression(b.Relationship.builtins["hash"], b.TupleArg(tup), res)

infix_ops = set(["+", "-", "*", "/", "^"])
infix_comps = set(["=", "!=", "<", "<=", ">", ">=", "implies"])
infixs = infix_ops.union(infix_comps)
prefix_ops = set(["abs", "log", "exp"])
agg_ops = set(["sum", "count", "min", "max", "all_different"])

# _variable_types = {
#     "cont": 40,
#     "int": 41,
#     "bin": 42,
# }

# _fo_operators = {
#     "+": 10,
#     "-": 11,
#     "*": 12,
#     "/": 13,
#     "^": 14,
#     "abs": 20,
#     "exp": 21,
#     "log": 22,
#     "range": 50,
# }

# _fo_comparisons = {
#     "=": 30,
#     "!=": 31,
#     "<=": 32,
#     ">=": 33,
#     "<": 34,
#     ">": 35,
#     "implies": 62,
# }

# _ho_operators = {
#     "sum": 80,
#     # "product":81,
#     "min": 82,
#     "max": 83,
#     "count": 84,
#     "all_different": 90,
# }



import warnings
from relationalai.semantics.sql import sql
from relationalai.semantics.sql.compiler import Compiler

__all__ = ["sql", "Compiler"]

warnings.warn(
    "relationalai.early_access.sql.* is deprecated. "
    "Please migrate to relationalai.semantics.sql.*",
    DeprecationWarning,
    stacklevel=2,
)


from abc import abstractmethod
from typing import Union

from relationalai.early_access.dsl.core.exprs.relational import RelationVariable
from relationalai.early_access.dsl.core.exprs.scalar import ScalarVariable
from relationalai.early_access.dsl.core.relations import Relation
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.schemas.exprs import Domain, Range
from relationalai.early_access.dsl.core.constraints.predicate.universal import RelationalEqualityConstraint
from relationalai.early_access.dsl.relations import EmptyRelation


#
# The abstract class SchemaComponent models a component that is introduced in the
# "declaration part" of a Schema
#
# Each declaration has the form:
#    v : t
# where v is a Var and t is a Type or Relation signature (arity and column types).
# The interpretation is that v is an element of Type t or has signature t.
#
class SchemaComponent:

    def __init__(self, name, dashes=0):
        self._name = name
        self._dashes:int = dashes
        self.before = None

    def basevar(self):
        if self.undecorated():
            return self

        if hasattr(self, 'before') and self.before is not None:
            return self.before.basevar()

        raise Exception("Need to extend SchemaComponent.basevar to create undecorated vars.")

    @abstractmethod
    def duplicate(self, name, t) -> 'SchemaComponent':
        pass

    @abstractmethod
    def relational(self) -> bool:
        pass

    @abstractmethod
    def typeof(self) -> Union[Type, Relation]:
        pass

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        return f"{self.name()} : {self.typeof().display()}"

    @abstractmethod
    def prefix(self, pref) -> 'SchemaComponent':
        pass

    def display_name(self):
        return self.name()

    def entityid(self):
        return hash(self.name())

    def decorated(self) -> bool:
        return self._dashes > 0

    def name(self):
        nm = self._name if isinstance(self._name, str) else self._name.display()
        return nm + self._dashes * "'"

    def sync_with(self, peer):
        return {self.name(): peer}

    def undash(self):
        if self._dashes > 0:
            return self.before
        return self

    def undecorated(self) -> bool:
        return self._dashes == 0


class RelationEmptyConstraint(RelationalEqualityConstraint):

    def __init__(self, left, sig):
        super().__init__(left, EmptyRelation(sig))


class RelationalComponent(RelationVariable, SchemaComponent):

    def __init__(self, name, t, dashes=0):
        super().__init__(name, dashes)
        self.relation = t
        if dashes < 2:
            self.after = RelationalComponent(name, t, dashes + 1)
            self.after.before = self

    def arity(self):
        return self.relation.arity()

    def display(self):
        return self.name()

    def domain(self):
        return Domain(self)

    def duplicate(self, name, t) -> 'RelationalComponent':
        return RelationalComponent(name, t, self._dashes)

    def empty(self):
        return RelationEmptyConstraint(self, self.relation)

    def entityid(self):
        return SchemaComponent.entityid(self)

    def prefix(self, pref):
        comp = RelationalComponent(prefix_concat(pref, self._name), self.relation)
        for i in range(self._dashes):
            comp = comp.after
        return comp

    def range(self):
        return Range(self)

    def rename(self, renaming):
        vname = self.display()
        if vname in renaming:
            var = renaming[vname]
            if isinstance(var, str):
                return RelationalComponent(var, self.relation, self._dashes)
            return var
        else:
            return self

    def refersto(self, varname: str):
        return self.display() == varname

    def relational(self) -> bool:
        return True

    def typeof(self):
        return self.relation

    def substitute(self, bindings):
        vname = self.display()
        if vname in bindings:
            return bindings[vname]
        else:
            return self


class ScalarComponent(ScalarVariable, SchemaComponent):

    def __init__(self, name, t, dashes=0):
        super().__init__(name, dashes)
        self._scalartype = t
        if dashes < 2:
            newcomp = ScalarComponent(name, t, dashes + 1)
            newcomp.before = self
            self.after = newcomp

    def entityid(self):
        return SchemaComponent.entityid(self)

    def prefix(self, pref):
        comp = ScalarComponent(prefix_concat(pref, self._name), self._scalartype)
        for i in range(self._dashes):
            comp = comp.after
        return comp

    def physical_typeof(self):
        return self._scalartype.root_unconstrained_type()

    def typeof(self):
        return self._scalartype

    def display(self):
        return self.name()

    def duplicate(self, name, t) -> 'ScalarComponent':
        return ScalarComponent(name, t, self._dashes)

    def rename(self, renaming):
        vname = self.display()
        if vname in renaming:
            var = renaming[vname]
            if isinstance(var, str):
                return ScalarComponent(var, self._scalartype, self._dashes)
            return var
        else:
            return self

    def simplify(self):
        return self


def prefix_concat(s1, s2):
    if not s1:
        return s2
    if not s2:
        return s1
    return s1 + s2[0].capitalize() + s2[1:]



from datetime import datetime, timezone

# Pick the right UTC tzinfo once at import time (this changed in 3.11+)
UTC = getattr(datetime, "UTC", timezone.utc)

def get_timestamp():
    return datetime.now(UTC)


"""
Helpers to generate Rel, many taken from the original PyRel implementation.
"""
from __future__ import annotations
import re
from relationalai.early_access.metamodel import ir, types, factory as f

# Rel primitive relations used by the compiler
# TODO: how to deal with varargs?
RelPrimitiveOuterJoin = f.relation("rel_primitive_outer_join", [])

# List of higher order relations in Rel
HIGHER_ORDER = set([
    "::std::common::count",
    "::std::common::sum",
    "::std::common::mean",
    "::std::common::max",
    "::std::common::min",
    "::std::common::sort",
    "::std::common::reverse_sort",
    "::std::common::top",
    "::std::common::bottom",
])

# Mappings of IR operators to the equivalent in Rel
OPERATORS = {
    # binary
    # TODO: maybe add the binary operators

    # aggregates
    "count": "pyrel_count",
    "sum": "::std::common::sum",
    "avg": "::std::common::mean",
    "max": "::std::common::max",
    "min": "::std::common::min",

    # std
    "range": "::std::common::range",
    "hash": "rel_primitive_hash_tuple_uint128",
    "uuid_to_string": "::std::common::uuid_string",

    # dates
    "date_year": "::std::common::date_year",
    "date_month": "::std::common::date_month",
    "date_day": "::std::common::date_day",
    "date_add": "::std::common::date_add",
    "dates_period_days": "pyrel_dates_period_days",
    "datetimes_period_milliseconds": "pyrel_datetimes_period_milliseconds",
    "datetime_add": "::std::common::datetime_add",
    "date_subtract": "::std::common::date_subtract",
    "datetime_subtract": "::std::common::datetime_subtract",
    "construct_datetime_ms_tz": "::std::common::^DateTime",
    "date_format": "::std::common::format_date",
    "datetime_format": "::std::common::format_datetime",
    "construct_date_from_datetime": "::std::common::^Date",

    # periods
    "millisecond": "::std::common::^Millisecond",
    "second": "::std::common::^Second",
    "minute": "::std::common::^Minute",
    "hour": "::std::common::^Hour",
    "day": "::std::common::^Day",
    "week": "::std::common::^Week",
    "month": "::std::common::^Month",
    "year": "::std::common::^Year",

    # math
    "abs": "::std::common::abs",
    "natural_log": "::std::common::natural_log",
    "sqrt": "std::common::sqrt",
    "maximum": "::std::common::maximum",
    "minimum": "::std::common::minimum",
    "//": "::std::common::trunc_divide",
    "isinf": "::std::common::Infinity",
    "isnan": "::std::common::NaN",
    "ceil": "::std::common::ceil",
    "floor": "::std::common::floor",

    # strings
    "string": "::std::common::string",
    "concat": "::std::common::concat",
    "starts_with": "::std::common::starts_with",
    "ends_with": "::std::common::ends_with",
    "contains": "::std::common::contains",
    "substring": "::std::common::substring",
    "num_chars": "pyrel_num_chars",
    "like_match": "::std::common::like_match",
    "lower": "::std::common::lowercase",
    "upper": "::std::common::uppercase",
    "strip": "rel_primitive_trim",

    # ints
    "parse_int64": "::std::common::parse_int",
    "parse_int128": "::std::common::parse_int128",

    # pragmas
    "rule_reasoner_sem_vo": "rel_primitive_force_faq_var_order",
    "rule_reasoner_phys_vo": "rel_primitive_force_var_order",
}

def rel_operator(ir_op):
    """ Maps an operator from the metamodel IR into the equivalent Rel operator. """
    if ir_op in OPERATORS:
        return OPERATORS[ir_op]
    else:
        return ir_op

def rel_typename(ir_type: ir.Type):
    """ Get the name of the type to use in Rel. """
    if isinstance(ir_type, ir.DecimalType):
        return f'pyrel_Decimal_{ir_type.precision}_{ir_type.scale}'
    if isinstance(ir_type, ir.ScalarType):
        if ir_type == types.Sha1:
            return 'Sha1'
        elif ir_type == types.Enum:
            return '::std::common::UInt128'
        elif ir_type == types.EntityTypeVar:
            return '::std::common::UInt128'
        elif ir_type == types.Hash:
            return "::std::common::UInt128"
        elif ir_type == types.UInt128:
            return '::std::common::UInt128'
        elif ir_type == types.RowId:
            return '::std::common::UInt128'
        elif ir_type == types.Bool:
            return '::std::common::Boolean'
        elif types.is_builtin(ir_type):
            return f'::std::common::{ir_type.name}'
        elif ir_type.super_types:
            # assuming the compiler narrowed this down somehow
            return rel_typename(ir_type.super_types.some())
        else:
            # user-defined types without a super type are user defined entities
            return '::std::common::UInt128'
    else:
        # TODO: how should we deal with Union types here?
        return '::std::common::Any'

# SEE: REL:      https://docs.relational.ai/rel/ref/lexical-symbols#keywords
#      CORE REL: https://docs.google.com/document/d/12LUQdRed7P5EqQI1D7AYG4Q5gno9uKqy32i3kvAWPCA
RESERVED_WORDS = [
    "and",
    "as",
    "bound",
    "declare",
    "def",
    "else",
    "end",
    "entity",
    "exists",
    "false",
    "for",
    "forall",
    "from",
    "ic",
    "if",
    "iff",
    "implies",
    "in",
    "module",
    "namespace",
    "not",
    "or",
    "requires",
    "then",
    "true",
    "use",
    "where",
    "with",
    "xor"
]

rel_sanitize_re = re.compile(r'[^\w:\[\]\^" ,]|^(?=\d)', re.UNICODE)
unsafe_symbol_pattern = re.compile(r"[^a-zA-Z0-9_]", re.UNICODE)

def sanitize(input_string, is_rel_name_or_symbol = False):
    # Replace non-alphanumeric characters with underscores
    if is_rel_name_or_symbol and "[" in input_string:
        string_parts = input_string.split('[', 1)
        sanitized_name_or_symbol = sanitize_identifier(string_parts[0])
        sanitized_rest = re.sub(rel_sanitize_re, "_", string_parts[1])
        sanitized = f"{sanitized_name_or_symbol}[{sanitized_rest}"
    else:
        if "::" in input_string: # TODO: This is a temp solution to avoid sanitizing the namespace
            sanitized = re.sub(rel_sanitize_re, "_", input_string)
        else:
            sanitized = sanitize_identifier(input_string)

    # Check if the resulting string is a keyword and append an underscore if it is
    if sanitized in RESERVED_WORDS:
        sanitized += "_"

    return sanitized

def sanitize_identifier(name: str) -> str:
    """
    Return a string safe to use as a top level identifier in rel, such as a variable or relation name.

    Args:
        name (str): The input identifier string.

    Returns:
        str: The sanitized identifier string.
    """

    if not name:
        return name

    safe_name = ''.join(c if c.isalnum() else '_' for c in name)
    if safe_name[0].isdigit():
        safe_name = '_' + safe_name
    if  safe_name in RESERVED_WORDS:
        safe_name = safe_name + "_" # preferring the pythonic pattern of `from_` vs `_from`
    return safe_name



from __future__ import annotations

from relationalai.early_access.metamodel import ir
from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel.visitor import Visitor, Rewriter
from relationalai.early_access.metamodel.util import OrderedSet
from relationalai.early_access.metamodel import helpers, factory as f, types, builtins
from typing import Optional, Any, Iterable
from collections import defaultdict

# Given an Output with a group of keys (some of them potentially null),
# * extract the lookups that bind (transitively) all the keys
# * extract the lookups that bind (transitively) properties of the keys
# * generate all the valid combinations of keys being present or not
#   * first all keys are present,
#   * then we remove one key at a time,
#   * then we remove two keys at a time,and so on.
#   * the last combination is when all the *nullable* keys are missing.
# * for each combination:
#   * create a compound (hash) key
#   * create a Logical that:
#       * contains all the relevant lookups for keys and properties
#       * contains negated lookups for null keys
#       * outputs using the compound key
#
# E.g., we go from
#
# Logical
#     Foo(foo)
#     rel1(foo, x)
#     Logical ^[v1=None]
#         rel2(foo, v1)
#     Logical ^[v2=None, k2=None]
#         rel3(foo, k2)
#         rel4(k2, v2)
#     Logical ^[v3=None, k3=None]
#         rel5(foo, y)
#         rel6(y, k3)
#         rel7(k3, v3)
#     output[foo, k2, k3](v1, v2, v3)
#
# to
#
# Logical
#     Logical
#         Logical
#             Foo(foo)
#             rel1(foo, x)
#             Logical ^[k2=None, k3=None]
#                 rel3(foo, k2)
#                 rel5(foo, y)
#                 rel6(y, k3)
#             Logical ^[v1=None]
#                 rel2(foo, v1)
#             Logical ^[v2=None, k2=None]
#                 rel4(k2, v2)
#             Logical ^[v3=None, k3=None]
#                 rel7(k3, v3)
#             construct(Hash, "Foo", foo, "Concept2", k2, "Concept3", k3, compound_key)
#             output[compound_key](v1, v2, v3)
#         Logical
#             Foo(foo)
#             rel1(foo, x)
#             Logical ^[k2=None, k3=None]
#                 rel5(foo, y)
#                 rel6(y, k3)
#             Not
#                 Logical
#                     rel3(foo, k2)
#             Logical ^[v1=None]
#                 rel2(foo, v1)
#             Logical ^[v3=None, k3=None]
#                 rel7(k3, v3)
#             construct(Hash, "Foo", foo, "Concept3", k3, compound_key)
#             output[compound_key](v1, None, v3)
#         Logical
#             Foo(foo)
#             rel1(foo, x)
#             Logical ^[k2=None, k3=None]
#                 rel3(foo, k2)
#                 rel5(foo, y)
#             Not
#                 Logical
#                     rel6(y, k3)
#             Logical ^[v1=None]
#                 rel2(foo, v1)
#             Logical ^[v2=None, k2=None]
#                 rel4(k2, v2)
#             construct(Hash, "Foo", foo, "Concept2", k2, compound_key)
#             output[compound_key](v1, v2, None)
#         Logical
#             Foo(foo)
#             rel1(foo, x)
#             Logical ^[k2=None, k3=None]
#                 rel5(foo, y)
#             Not
#                 Logical
#                     rel3(foo, k2)
#                     rel6(y, k3)
#             Logical ^[v1=None]
#                 rel2(foo, v1)
#             construct(Hash, "Foo", foo, compound_key)
#             output[compound_key](v1, None, None)

class ExtractKeys(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        visitor = IdentifyKeysVisitor()
        model.accept(visitor)
        return ExtractKeysRewriter(visitor).walk(model)

class ExtractInfo:
    def __init__(self, vars_to_extract: OrderedSet[ir.Var]):
        self.original_keys = vars_to_extract
        # the subsets of the original keys that are nullable or not.
        self.non_nullable_keys: OrderedSet[ir.Var] = OrderedSet()
        self.nullable_keys: OrderedSet[ir.Var] = OrderedSet()

        # lookup tasks that transitively bind all the key vars
        # e.g., if the key is Z, in foo(X, Y), bar(Y, Z), baz(Z, W)
        # we extract foo(X, Y), bar(Y, Z)
        self.key_lookups: list[ir.Task] = []

        # lookup tasks that transitively bind all the properties of the keys
        # e.g., if the key is Z, in bar(Y, Z), baz(Z, W), qux(W, V)
        # we extract baz(Z, W), qux(W, V).
        # store the lookups by key. each key might bind multiple properties, each list in
        # dict keeps track of one of these properties
        self.inner_property_lookups: dict[ir.Var, list[OrderedSet[ir.Task]]] = defaultdict(list)

        # other top-level tasks that are not key lookups. These need to be kept in each
        # match case of the rewritten Logical.
        self.non_key_top_level_tasks: list[ir.Task] = []

        # the original output task. New outputs are generated based on the keys and
        # properties in the original output
        self.original_output: Optional[ir.Output] = None

        # variables coming from concept lookups, that are not nullable
        # this includes lookups at the top-level, as well as inside logicals when
        # hoisted without a None default.
        self.non_nullable_vars: OrderedSet[ir.Var] = OrderedSet()

class IdentifyKeysVisitor(Visitor):
    def __init__(self):
        self.extract_info_for_logical: dict[ir.Logical, ExtractInfo] = {}
        self.curr_info = None

    def enter(self, node: ir.Node, parent: Optional[ir.Node]=None) -> Visitor:
        if isinstance(node, ir.Logical):
            outputs = [x for x in node.body if isinstance(x, ir.Output) and x.keys]
            if not outputs:
                return self
            assert len(outputs) == 1, "multiple outputs with keys in a logical"
            if not outputs[0].keys:
                return self

            # Logical with an output that has keys
            info = ExtractInfo(OrderedSet.from_iterable(outputs[0].keys))

            # Set the original output
            info.original_output = outputs[0]

            # The original keys and any intermediate vars needed to correctly bind the keys.
            # Any transitive key vars are added during the `find_key_lookups_fixpoint` calls
            extended_keys:OrderedSet[ir.Var] = OrderedSet.from_iterable(outputs[0].keys)

            # First, collect all the top-level lookups
            top_level_lookups = []
            for task in node.body:
                if isinstance(task, ir.Lookup):
                    top_level_lookups.append(task)

            # Find all the key lookups at the top level. Property lookups are handled later
            # and stored in non_key_top_level_tasks
            key_lookups = self.find_key_lookups_fixpoint(top_level_lookups, extended_keys)

            # Then, deal with lookups inside logicals (with hoisted defaults)
            for task in node.body:
                processed = False

                # Top-level concept lookups identify variables that are not nullable
                if isinstance(task, ir.Lookup) and helpers.is_concept_lookup(task):
                    vars = helpers.vars(task.args)
                    if vars[0] not in info.non_nullable_vars:
                        info.non_nullable_vars.add(vars[0])

                if isinstance(task, ir.Logical):
                    for h in task.hoisted:
                        # Hoisted vars without a default are not nullable
                        if isinstance(h, ir.Var):
                            info.non_nullable_vars.add(h)
                        elif isinstance(h, ir.Default):
                            # Hoisted vars with a non-None default are not nullable
                            if h.value is not None:
                                info.non_nullable_vars.add(h.var)
                            elif h.var in info.original_keys and h.var not in info.non_nullable_vars:
                                info.nullable_keys.add(h.var)

                    # Find key lookups inside the logical
                    current_lookups = self.find_key_lookups_fixpoint(task.body, extended_keys)
                    key_lookups.update(current_lookups)

                    # Find property lookups inside the logical. Assume that each logical
                    # binds a separate property for each key
                    current_property_lookups = self.find_property_lookups_fixpoint(task.body, extended_keys)
                    for k, v in current_property_lookups.items():
                        info.inner_property_lookups[k].append(v)

                    if current_lookups or current_property_lookups:
                        processed = True

                if task in key_lookups:
                    continue

                # If the task does not contain key or property lookups, then we need to
                # maintain these at the top level during the actual rewrite
                if not processed and \
                    (
                        isinstance(task, ir.Logical) or \
                        isinstance(task, ir.Lookup) or \
                        isinstance(task, ir.Construct) or \
                        isinstance(task, ir.Aggregate) or \
                        isinstance(task, ir.Rank)
                    ):
                    info.non_key_top_level_tasks.append(task)

            info.non_nullable_keys = info.original_keys - info.nullable_keys
            info.key_lookups = list(key_lookups)

            # we only need to transform the logical if there are nullable keys
            if info.nullable_keys:
                self.extract_info_for_logical[node] = info
                self.curr_info = info
        return self

    def leave(self, node: ir.Node, parent: Optional[ir.Node]=None) -> ir.Node:
        if not self.curr_info:
            return node

        if isinstance(node, ir.Aggregate):
            # we assume that variables appearing in aggregate group-by's are not nullable
            for v in node.group:
                if v in self.curr_info.nullable_keys:
                    self.curr_info.nullable_keys.remove(v)
        elif isinstance(node, ir.Logical) and node in self.extract_info_for_logical:
            # if the set of nullable keys became empty, we shouldn't attempt to transform the logical
            if not self.curr_info.nullable_keys:
                self.extract_info_for_logical.pop(node)
            self.curr_info = None

        return node

    def find_key_lookups_fixpoint(self, tasks:Iterable[ir.Task], keys:OrderedSet[ir.Var]):
        # lookups with a single argument correspond to concepts.
        # we should keep them ahead of the other lookups.
        concept_lookups = OrderedSet()
        # for lookups with multiple arguments, we start from those that have a key as the last
        # argument and move backwards. that's why each time we insert at the front of the list
        lookups = OrderedSet()

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in tasks:
                if isinstance(task, ir.Lookup) and task not in lookups and task not in concept_lookups:
                    vars = helpers.vars(task.args)
                    if len(vars) == 1 and vars[0] in keys:
                        concept_lookups.add(task)
                        there_is_progress = True
                    elif len(vars) > 1 and all(v in keys for v in vars[1:]):
                        assert isinstance(vars[0], ir.Var)
                        keys.add(vars[0])
                        lookups.prepend(task)
                        there_is_progress = True

        return concept_lookups | lookups

    def find_property_lookups_fixpoint(self, tasks:Iterable[ir.Task], keys:Iterable[ir.Var]):
        property_lookups = defaultdict(OrderedSet)

        # Keep track of transitive vars seen for each key
        seen_vars: dict[ir.Var, set[ir.Var]] = {k: set([k]) for k in keys}

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in tasks:
                if isinstance(task, ir.Lookup):
                    vars = task.args
                    if len(vars) <= 1:
                        # Singleton lookups cannot be for properties
                        continue
                    if any(task in v for v in property_lookups.values()):
                        # Already seen this lookup - don't add it again
                        continue
                    for (var, inter_vars) in seen_vars.items():
                        if any(nv in vars for nv in inter_vars):
                            # If there are any other keys in the lookup, then it is not
                            # purely a lookup of this key's properties
                            #
                            # TODO: handle the case where there are multiple keys for a
                            # single property.
                            if any(isinstance(v, ir.Var) and v in keys and v != var for v in vars):
                                continue

                            # If any seen var is in the args
                            property_lookups[var].add(task)
                            seen_vars[var].update([v for v in vars if isinstance(v, ir.Var)])
                            there_is_progress = True
                elif isinstance(task, ir.Construct):
                    if any(task in v for v in property_lookups.values()):
                        continue
                    for (var, inter_vars) in seen_vars.items():
                        if task.id_var in inter_vars:
                            property_lookups[var].add(task)
                            there_is_progress = True

        return property_lookups

class ExtractKeysRewriter(Rewriter):
    def __init__(self, visitor: IdentifyKeysVisitor):
        super().__init__()
        self.visitor = visitor

    def handle_logical(self, node: ir.Logical, parent: ir.Node, ctx:Optional[Any]=None) -> ir.Logical:
        new_body = self.walk_list(node.body, node)

        # We are in a logical with an output at this level.
        if node in self.visitor.extract_info_for_logical:
            info = self.visitor.extract_info_for_logical[node]

            # Update the key sets based on the identified non-nullable vars
            # remove nullable keys that were inferred to be non-nullable,
            # and add them to the non-nullable set
            non_null_nullable_keys = info.nullable_keys & info.non_nullable_vars
            info.non_nullable_keys.update(non_null_nullable_keys)
            info.nullable_keys = info.nullable_keys - info.non_nullable_vars

            # Get the key lookups. Each output case will start with these lookups
            top_level_key_lookups = list(info.key_lookups)

            # Create a compound key that will be used in place of the original keys.
            compound_key = f.var("compound_key", types.Hash)

            # Get all cases for each combination of null/non-null keys
            match_cases = self._nullable_key_combinations(info, top_level_key_lookups, compound_key)

            return f.logical(tuple(match_cases), [])
        else:
            return node if new_body is node.body else f.logical(new_body, node.hoisted)

    # Generate all the combinations of nullable keys being present or not.
    # For each such combination, generate a list of lookups that will be used in the Match.
    # In total, return a list of lists of lookups, covering all the Match cases.
    def _nullable_key_combinations(
            self,
            info:ExtractInfo,
            inner_key_lookups:list[ir.Task],
            compound_key:ir.Var):
        return self._nullable_key_combinations_rec(info, [], 0, inner_key_lookups, compound_key)

    def _nullable_key_combinations_rec(
            self,
            info: ExtractInfo,
            nullable_non_null_keys: list[ir.Var],
            idx: int,
            inner_key_lookups: list[ir.Task],
            compound_key:ir.Var):

        if idx < len(info.nullable_keys):
            key = info.nullable_keys[idx]
            case1_list = self._nullable_key_combinations_rec(info, nullable_non_null_keys + [key], idx + 1, inner_key_lookups, compound_key)
            case2_list = self._nullable_key_combinations_rec(info, nullable_non_null_keys, idx + 1, inner_key_lookups, compound_key)
            case1_list.extend(case2_list)
            return case1_list
        else:
            # Create a copy to mutate
            inner_key_lookups_ = list(inner_key_lookups)

            # Make sure the orig output is set
            assert isinstance(info.original_output, ir.Output)

            curr_non_null_keys = OrderedSet[ir.Var]()
            curr_null_keys = OrderedSet[ir.Var]()
            for key in info.original_keys:
                if key in info.non_nullable_keys or key in nullable_non_null_keys:
                    curr_non_null_keys.add(key)
                else:
                    curr_null_keys.add(key)

            removed_lookups = self._remove_from_key_lookups(inner_key_lookups_, curr_null_keys, curr_non_null_keys)

            # Build the inner lookups.
            # Start with the non-null key lookups. Wrap these in a logical where the non-null
            # keys are hoisted
            inner_lookups: list[ir.Task] = [f.logical(inner_key_lookups_, list(curr_non_null_keys))]

            # Add other top-level tasks that are not key or property lookups
            inner_lookups.extend(info.non_key_top_level_tasks)

            # Add negated lookups for each null key
            if removed_lookups:
                for (_, removed) in removed_lookups.items():
                    if len(removed) == 1:
                        negated_null_key_lookups = f.not_(removed[0])
                    else:
                        negated_null_key_lookups = f.not_(f.logical(removed, []))
                    inner_lookups.append(negated_null_key_lookups)

            # Add property lookups for each non-null key
            for key in curr_non_null_keys:
                for tasks in info.inner_property_lookups.get(key, []):
                    # Create a separate logical for each key's property lookups. This is
                    # needed so that flatten can split the dependencies correctly
                    prop_lookups = []
                    hoisted_prop_vars = OrderedSet()
                    for task in tasks:
                        if task not in prop_lookups:
                            prop_lookups.append(task)
                            assert isinstance(task, ir.Lookup)

                            # only care about hoisting properties that are in the output
                            vars = helpers.vars(task.args)
                            output_vars = helpers.output_vars(info.original_output.aliases)
                            hoisted_prop_vars.update([v for v in vars if v in output_vars])
                    if prop_lookups:
                        hoisted_props = [ir.Default(v, None) for v in hoisted_prop_vars]
                        inner_lookups.append(f.logical(prop_lookups, hoisted_props))

            # Create the arguments for constructing the compound key. Only include the
            # values for non-null keys. Null key names are still included.
            values: list[ir.Value] = [compound_key.type]
            for key in info.original_keys:
                assert isinstance(key.type, ir.ScalarType)
                values.append(ir.Literal(types.String, key.type.name))
                if key in curr_non_null_keys:
                    values.append(key)
            inner_lookups.append(ir.Construct(
                None,
                tuple(values),
                compound_key,
                OrderedSet().frozen()
            ))

            # Add the output for this combination of null/non-null keys.
            # First, get the annos
            annos = list(info.original_output.annotations)

            if info.original_output.keys:
                annos.append(f.annotation(builtins.output_keys, tuple(info.original_output.keys)))

            # Then, construct the output aliases. Drop the arguments that are either
            # properties of currently null keys or the null keys themselves
            curr_null_keys_or_properties = OrderedSet[ir.Var]()
            output_vars = helpers.output_vars(info.original_output.aliases)
            for key in curr_null_keys:
                for tasks in info.inner_property_lookups.get(key, []):
                    for task in tasks:
                        assert isinstance(task, ir.Lookup)
                        vars = helpers.vars(task.args)
                        curr_null_keys_or_properties.update([v for v in vars if v in output_vars])
                if key in output_vars:
                    curr_null_keys_or_properties.add(key)

            new_output_aliases = []
            for (alias, var) in info.original_output.aliases:
                # Need a placeholder (None) for this output var so that columns still
                # line up. This will be excluded later during Flatten so we don't
                # actually output None values.
                out_var = None if var in curr_null_keys_or_properties else var
                new_output_aliases.append((alias, out_var))
            inner_lookups.append(f.output(new_output_aliases, [compound_key], annos=annos))

            return [f.logical(inner_lookups, [])]

    # Null keys should not be looked up. Non-nullable keys should also not be looked up,
    # since they will be looked up separately in their own logical.
    #
    # Returns the removed lookups as a dict mapping from key to lookups
    def _remove_from_key_lookups(self, lookups: list[ir.Task], vars_to_purge: OrderedSet[ir.Var], non_nullable_keys: OrderedSet[ir.Var]):
        removed = defaultdict(list)
        # keep track of transitive keys too
        keys = [[v] for v in vars_to_purge]

        there_is_progress = True
        while there_is_progress:
            there_is_progress = False
            for task in lookups:
                assert isinstance(task, ir.Lookup)
                vars = helpers.vars(task.args)
                for ks in keys:
                    if vars[-1] in ks:
                        lookups.remove(task)
                        # ks[0] is the original key that led to this chain of lookups
                        removed[ks[0]].append(task)
                        new_vars = [v for v in vars if v not in vars_to_purge and v not in non_nullable_keys]
                        ks.extend(new_vars)
                        there_is_progress = True
                        break

        return removed



__version__ = "0.0.0"

from .core import Graph

import warnings

warnings.warn(
    (
        "\n\nIf you are using this package, please contact the symbolic reasoning team.\n\n"
        "This library is still in the early stages of development. Interfaces "
        "_will_ change. Performance is appropriate only for exploring small "
        "graphs; there are known asymptotic catastrophes. For more, please see the README."
    ),
    FutureWarning,
    stacklevel=2
)

__all__ = [
    "Graph",
]



# For builder components.
from relationalai.early_access.builder import Integer
from relationalai.early_access.builder import define, max, sum, not_

from relationalai.early_access.paths.graph import Graph
from relationalai.early_access.paths.path_algorithms.one_sided_ball_upto import ball_upto
from relationalai.early_access.paths.path_algorithms.one_sided_ball_repetition import ball_with_repetition


def compute_usp(g: Graph, Source, Target, max_length = None):
    # Computes the Union of Shortest Paths (USP) from a source set to a destination set
    # USP is a subgraph of the original graph

    edge = g.Edge
    Node = g.Node

    n, d = Integer.ref(), Integer.ref()
    tgt, u, v, w = Node.ref(), g.Node.ref(), g.Node.ref(), g.Node.ref()

    if max_length is None:
        ball = ball_upto(g, Source, Target)
        usp_nodes = g.model.Relationship(f"Node {{{Node}}} is in the USP")
        usp = g.model.Relationship(f"Edge {{{Node}}} to {{{Node}}} is in the USP")
        boundary = g.model.Relationship(f"Target node {{{Node}}} is in the boundary of the ball")

        # propagate backwards to find the nodes that are in the USP:
        define(boundary(tgt)).where(
            Target(tgt),
            ball(n, tgt)
        )

        define(usp_nodes(tgt)).where(
            boundary(tgt)
        )

        define(usp_nodes(u)).where(
            usp(u, v)
        )

        define(usp(u, v)).where(
            usp_nodes(v),
            ball(d, v),
            ball(d - 1, u),
            edge(u, v)
        )
    else:
        ball = ball_with_repetition(g, Source, max_length)
        max_dist = g.model.Relationship("The maximum distance of a target node form the source nodes is {Integer}")
        usp = g.model.Relationship(f"{{{Node}}} at distance {{Integer}} from the source nodes is connected with {{{Node}}}")
        boundary = g.model.Relationship(f"Target pair {{{Node}}} {{Integer}} is in the boundary of the ball")

        define(max_dist(max(n))).where(
            ball(n, tgt),
            Target(tgt)
        )

        define(boundary(tgt, n)).where(
            max_dist(n),
            Target(tgt),
            ball(n, tgt)
        )

        define(usp(u, n, tgt)).where(
            boundary(tgt, n + 1),
            ball(n, u),
            edge(u, v)
        )

        define(usp(u, n, v)).where(
            usp(v, n + 1, w),
            ball(n, u),
            edge(u, v)
        )


    return usp, boundary




def compute_nsp_from_usp(g: Graph, usp, Source, Target, Boundary):
    Node = g.Node

    n, m = Integer.ref(), Integer.ref()
    tgt, u, v = Node.ref(), Node.ref(), Node.ref()

    if usp._arity() == 2:
        nsp = g.model.Relationship(f"Number of shortest paths from {{{Node}}} to destination is {{Integer}}")

        define(nsp(tgt, 1)).where(
            Boundary(tgt)
        )

        define(nsp(tgt, 1)).where(
            Target(tgt),
            Source(tgt)
        )

        define(nsp(u, sum(v, n).per(u))).where(
            nsp(v, n),
            usp(u, v)
        )
    else:
        nsp = g.model.Relationship(f"Number of shortest paths from pair {{{Node}}} {{Integer}} to destination is {{Integer}}")

        define(nsp(tgt, n, 1)).where(
            Boundary(tgt, n)
        )

        define(nsp(u, n, sum(v, n + 1, m).per(u, n))).where(
            nsp(v, n + 1, m),
            usp(u, n, v),
            not_(Target(u))
        )

        define(nsp(u, n, 1 + sum(v, n + 1, m).per(u, n))).where(
            nsp(v, n + 1, m),
            usp(u, n, v),
            Target(u),
            not_(Boundary(u, n))
        )

    return nsp


def compute_nsp(g: Graph, Source, Target, max_length = None):
    # Computes the number of shortest paths (NSP) from a source set to a destination set

    usp, Boundary  = compute_usp(g, Source, Target, max_length)
    nsp = compute_nsp_from_usp(g, usp, Source, Target, Boundary)

    return nsp



from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers, types
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.lqp import ir as lqp, utils, types as lqp_types
from relationalai.early_access.lqp.primitives import (
    relname_to_lqp_name, lqp_operator, lqp_avg_op, is_monotype, primitive_types
)
from relationalai.early_access.lqp.pragmas import pragma_to_lqp_name
from relationalai.early_access.lqp.types import meta_type_to_lqp
from relationalai.early_access.lqp.constructors import (
    mk_abstraction, mk_and, mk_exists, mk_or, mk_primitive, mk_type, mk_value, mk_var,
    mk_pragma
)
from relationalai.early_access.lqp.utils import TranslationCtx, gen_unique_var
from relationalai.early_access.lqp.validators import assert_valid_input

from decimal import Decimal as PyDecimal
from datetime import datetime, date
from typing import Tuple, cast, Union, Optional
import re

""" Main access point. Converts the model IR to an LQP transaction. """
def to_lqp(model: ir.Model, fragment_name: bytes) -> tuple[Optional[tuple], lqp.Transaction]:
    assert_valid_input(model)
    ctx = TranslationCtx(model)
    decls: list[lqp.Declaration] = []
    reads: list[lqp.Read] = []

    # Add pyrel error attributes to reads.
    for err_id in _extract_pyrel_error_ids(ctx, model):
        ctx.output_ids.append(err_id)

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        new_decls = _translate_to_decls(ctx, subtask)
        decls.extend(new_decls)

    reads.extend(_get_output_reads(ctx.output_ids))

    export_info = None
    if len(ctx.export_ids) > 0:
        export_filename, col_types, export_reads = _get_export_reads(ctx.export_ids)
        reads.extend(export_reads)
        export_info = (export_filename, col_types)

    debug_info = lqp.DebugInfo(id_to_orig_name=ctx.rel_id_to_orig_name, meta=None)
    fragment_id = lqp.FragmentId(id=fragment_name, meta=None)
    fragment = lqp.Fragment(id=fragment_id, declarations=decls, meta=None, debug_info=debug_info)
    define_op = lqp.Define(fragment=fragment, meta=None)

    txn = lqp.Transaction(
        epochs=[
            lqp.Epoch(
                reads=reads,
                writes=[lqp.Write(write_type=define_op, meta=None)],
                meta=None
            )
        ],
        configure=lqp.construct_configure({}, None),
        meta=None,
    )

    lqp.validate_lqp(txn)
    return (export_info, txn)

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # Unions may not return anything. The generated IR contains a None value when this
        # happens. We ignore it here.
        return [v for v in helpers.output_values(effect.aliases) if v]
    else:
        return list(effect.args)

def _get_output_reads(output_ids: list[tuple[lqp.RelationId, str]]) -> list[lqp.Read]:
    reads = []
    for (rel_id, name) in output_ids:
        assert isinstance(rel_id, lqp.RelationId)
        output = lqp.Output(name=name, relation_id=rel_id, meta=None)
        reads.append(lqp.Read(read_type=output, meta=None))
    return reads

def _get_export_reads(export_ids: list[tuple[lqp.RelationId, int, lqp.Type]]) -> tuple[str, list, list[lqp.Read]]:
    reads = []
    csv_columns = []
    col_info = []
    for (col_id, col_num, col_type) in sorted(export_ids, key=lambda x: x[1]):
        col_name = f"col{col_num:03}"
        csv_columns.append(lqp.ExportCSVColumn(column_name=col_name, column_data=col_id, meta=None))
        col_info.append((col_name, col_type))

    # Generate a hashed name based on relation ids
    rel_ids = tuple([id[0] for id in export_ids])
    export_filename = "export_" + str(abs(hash(rel_ids)))

    # Note that the engine will append the transaction id to the export path for access
    # control reasons. So the actual final path will be
    # `.../export/{export_filename}/data_{txn_id}`
    export_path = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/export/{export_filename}/data"
    export_csv_config = lqp.ExportCSVConfig(
        path=export_path,
        data_columns=csv_columns,
        compression="gzip",
        partition_size=200,
        meta=None,
    )
    reads.append(lqp.Read(read_type=lqp.Export(config=export_csv_config, meta=None), meta=None))
    return (export_filename, col_info, reads)

def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    effects = collect_by_type((ir.Output, ir.Update), rule)
    aggregates = collect_by_type(ir.Aggregate, rule)
    ranks = collect_by_type(ir.Rank, rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return []

    assert len(ranks) == 0 or len(aggregates) == 0, "rules cannot have both aggregates and ranks"

    conjuncts = []
    for task in rule.body:
        if isinstance(task, (ir.Output, ir.Update)):
            continue
        conjuncts.append(_translate_to_formula(ctx, task))

    # Aggregates reduce over the body
    if aggregates or ranks:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            conjuncts.append(_translate_aggregate(ctx, aggr, aggr_body))
        for rank in ranks:
            conjuncts.append(_translate_rank(ctx, rank, aggr_body))

    return [_translate_effect(ctx, effect, mk_and(conjuncts)) for effect in effects]

# Translates an effect (export, output, or update) into the corresponding def. Note that
# this method only generates the def and not the LQP read operation, which is added later by
# `_get_output_reads` and `_get_export_reads`.
def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp.Formula) -> lqp.Declaration:
    bindings = _effect_bindings(effect)

    def _is_export(e):
        return isinstance(e, ir.Output) and rel_builtins.export_annotation in e.annotations

    if isinstance(effect, ir.Output):
        projection, eqs, suffix = _translate_output_bindings(ctx, bindings)
        meta_id = effect.id

        if _is_export(effect):
            def_name = "export_relation" + suffix
        else:
            def_name = "output" + suffix

        # Uniquify the output name
        def_name = ctx.def_names.get_name_by_id(meta_id, def_name)
    else:
        projection, eqs = _translate_bindings(ctx, bindings)
        meta_id = effect.relation.id
        def_name = effect.relation.name

    eqs.append(body)
    new_body = mk_and(eqs)

    rel_id = get_relation_id(ctx, def_name, meta_id)

    # Context bookkeeping for exports and outputs
    if _is_export(effect):
        # The row id is the first n-1 elements, and the actual data is the last element. Its
        # type is stored in the first element of the tuple.
        col_type = projection[-1][1]
        _col_num_match = re.search(r"export_relation_col([0-9]+)", def_name)
        assert _col_num_match, f"Could not find column number in suffix: {def_name}"
        col_num = int(_col_num_match.group(1))
        ctx.export_ids.append((rel_id, col_num, col_type))
    elif isinstance(effect, ir.Output):
        ctx.output_ids.append((rel_id, def_name))

    return lqp.Def(
        name = rel_id,
        body = mk_abstraction(projection, new_body),
        attrs = [],
        meta = None,
    )

def _translate_output_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula], str]:
    symbol_literals = []
    non_symbols = []
    for binding in bindings:
        if isinstance(binding, ir.Literal) and binding.type == types.Symbol:
            symbol_literals.append(binding.value)
        else:
            non_symbols.append(binding)
    projection, eqs = _translate_bindings(ctx, non_symbols)
    if len(symbol_literals) > 0:
        name_suffix = "_"
        name_suffix += "_".join(symbol_literals)
    else:
        name_suffix = ""

    return projection, eqs, name_suffix

def _translate_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.Formula) -> lqp.Formula:
    # Ascending rank is constructed using rel_primitive_sort. If a limit is added to an
    # ascending rank we can use rel_primitive_top for an efficient evaluation.
    #
    # Descending rank is constructed as an ascending sort, then the ascending rank is
    # subtracted from a count of the elements (plus 1 so that we still start at 1).
    # Adding a limit to a descending rank is done by adding a filter of rank <= limit.

    # Limits are the sort plus a filter on rank <= limit.
    if all(o for o in rank.arg_is_ascending):
        ascending = True
    elif all(not o for o in rank.arg_is_ascending):
        ascending = False
    else:
        raise Exception("Mixed orderings in rank are not supported yet.")

    # Filter out the group-by variables, since they are introduced outside the rank.
    input_args, input_eqs = _translate_bindings(ctx, list(rank.args))
    introduced_meta_projs = [arg for arg in rank.projection if arg not in rank.group and arg not in rank.args]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))

    # rank expects an Int128 result, but the primitive will return an Int64 result.
    # we need to set up an intermediary variable to hold the Int64 result, and a cast
    # to convert it to Int128.
    result_var, _ = _translate_term(ctx, rank.result)
    # The primitive will return an Int64 result, so we need a var to hold the intermediary.
    result_64_var = gen_unique_var(ctx, "result_64")
    result_64_type = mk_type(lqp.TypeName.INT)

    cast = lqp.Cast(input=result_64_var, result=result_var, meta=None)

    body = mk_and([body] + input_eqs + projected_eqs)
    abstr_args = input_args + projected_args

    if ascending:
        ranker = _translate_ascending_rank(ctx, rank.limit, result_64_var, body, abstr_args)
    else:
        ranker = _translate_descending_rank(ctx, rank.limit, result_64_var, body, abstr_args)

    return mk_exists([(result_64_var, result_64_type)], mk_and([ranker, cast]))

def _translate_descending_rank(ctx: TranslationCtx, limit: int, result: lqp.Var, body: lqp.Formula, abstr_args) -> lqp.Formula:
    result_var = result
    result_type = mk_type(lqp.TypeName.INT)

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]

    # Construct a conjunction of the ranking, a counter for the body, a subtraction
    # of the rank from the count and an addition of 1. Wrap this in an abstraction.
    count_res = gen_unique_var(ctx, "count_res")

    # Add one to the count to account for the rank starting at 1.
    one, one_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
    one_bigger = gen_unique_var(ctx, "one_bigger")
    addition = mk_primitive("rel_primitive_add_monotype", [count_res, one, one_bigger])

    # Subtract the rank from the count + 1
    asc_rank = gen_unique_var(ctx, "asc_rank")
    subtraction = mk_primitive("rel_primitive_subtract_monotype", [one_bigger, result_var, asc_rank])

    # Construct the ranking
    desc_ranking_terms = [asc_rank] + [v[0] for v in abstr_args]
    ranking = lqp.FFI(
        meta=None,
        name="rel_primitive_sort",
        args=[mk_abstraction(new_abstr_args, body)],
        terms=desc_ranking_terms,
    )

    # Count the number of rows in the body
    count_type = meta_type_to_lqp(types.Int64)
    count_var, count_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "counter")
    desc_body = mk_and([body, count_eq])
    aggr_abstr_args = new_abstr_args + [(count_var, count_type)]
    count_aggr = lqp.Reduce(
        op=lqp_operator(
            ctx.var_names,
            "count",
            "count",
            mk_type(lqp.TypeName.INT)
        ),
        body=mk_abstraction(aggr_abstr_args, desc_body),
        terms=[count_res],
        meta=None
    )

    # Bring it all together and do the maths.
    ranking = mk_exists(
        vars=[
            (asc_rank, result_type),
            (count_res, result_type),
            (one, result_type),
            (one_bigger, result_type)
        ],
        value=mk_and([ranking, count_aggr, one_eq, addition, subtraction])
    )

    # If there is a limit, we need to add a filter to the ranking.
    # Wrap with a rank <= limit
    if limit != 0:
        limit_term, _ = _translate_term(ctx, ir.Literal(types.Int64, limit))
        limiter = mk_primitive("rel_primitive_lt_eq_monotype", [result_var, limit_term])
        ranking = mk_and([ranking, limiter])

    return ranking

def _translate_ascending_rank(ctx: TranslationCtx, limit: int, result_var: lqp.Var, body: lqp.Formula, abstr_args) -> lqp.Formula:
    terms = [result_var] + [v[0] for v in abstr_args]

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]
    sort_abstr = mk_abstraction(new_abstr_args, body)

    if limit == 0:
        return lqp.FFI(
            meta=None,
            name="rel_primitive_sort",
            args=[sort_abstr],
            terms=terms,
        )
    else:
        limit_type = meta_type_to_lqp(types.Int64)
        limit_var, limit_eq = constant_to_var(ctx, to_lqp_value(limit, types.Int64), "limit")
        limit_abstr = mk_abstraction([(limit_var, limit_type)], limit_eq)
        return lqp.FFI(
            meta=None,
            name="rel_primitive_top",
            args=[sort_abstr, limit_abstr],
            terms=terms,
        )

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula) -> Union[lqp.Reduce, lqp.Formula]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "avg", "min", "max", "rel_primitive_solverlib_ho_appl")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    meta_output_terms = []
    meta_input_terms = []

    for (field, arg) in zip(aggr.aggregation.fields, aggr.args):
        if field.input:
            meta_input_terms.append(arg)
        else:
            meta_output_terms.append(arg)

    output_terms = [_translate_term(ctx, term) for term in meta_output_terms]
    output_vars = [term[0] for term in output_terms]

    body_conjs = [body]
    input_args, input_eqs = _translate_bindings(ctx, meta_input_terms)

    # Filter out the group-by variables, since they are introduced outside the aggregation.
    # Input terms are added later below.
    introduced_meta_projs = [arg for arg in aggr.projection if arg not in aggr.group and arg not in meta_input_terms]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    abstr_args: list[Tuple[lqp.Var, lqp.Type]] = projected_args + input_args

    if aggr_name == "count":
        assert len(output_terms) == 1, "Count and avg expect a single output variable"
        assert isinstance(meta_output_terms[0], ir.Var)
        # Count sums up "1" for each row. We use the expected output type for the type
        # of the count variable.
        typ = meta_type_to_lqp(meta_output_terms[0].type)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, meta_output_terms[0].type), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, typ))
        body = mk_and(body_conjs)

    # Average needs to wrap the reduce in Exists(Conjunction(Reduce, div))
    if aggr_name == "avg":
        assert len(output_vars) == 1, "avg should only have one output variable"
        output_var = output_vars[0]

        # Count sums up "1" for each row. We make the reasonably safe assumption that there
        # are less than 2^31 rows in the body.
        count_type = meta_type_to_lqp(types.Int64)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, count_type))
        body = mk_and(body_conjs)

        # The average will produce two output variables: sum and count.
        sum_result = gen_unique_var(ctx, "sum")
        count_result = gen_unique_var(ctx, "count")

        # Second to last is the variable we're summing over.
        (sum_var, sum_type) = abstr_args[-2]

        result = lqp.Reduce(
            op=lqp_avg_op(ctx.var_names, aggr.aggregation.name, sum_var.name, sum_type),
            body=mk_abstraction(abstr_args, body),
            terms=[sum_result, count_result],
            meta=None,
        )

        if sum_type == count_type:
            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_result, output_var])
            conjunction = mk_and([result, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type)]
        else:
            # If the sum type and count type don't match, we need to cast the count
            count_casted = gen_unique_var(ctx, "count_casted")
            count_cast = lqp.Cast(input=count_result, result=count_casted, meta=None)

            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_casted, output_var])
            conjunction = mk_and([result, count_cast, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type), (count_casted, sum_type)]

        return mk_exists(result_terms, conjunction)

    # `input_args`` hold the types of the input arguments, but they may have been modified
    # if we're dealing with a count, so we use `abstr_args` to find the type.
    (aggr_arg, aggr_arg_type) = abstr_args[-1]
    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp.Reduce(
        op=lqp_operator(ctx.var_names, aggr.aggregation.name, aggr_arg.name, aggr_arg_type),
        body=mk_abstraction(abstr_args, body),
        terms=output_vars,
        meta=None
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(arg=_translate_to_formula(ctx, task.task), meta=None)
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = _translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "Construct should have at least one value"
        assert isinstance(task.values[0], ir.ScalarType), "Construct should start with a named ScalarType"
        name = task.values[0].name
        terms = [_translate_term(ctx, ir.Literal(types.String, name))]
        terms.extend([_translate_term(ctx, arg) for arg in task.values[1:]])
        terms.append(_translate_term(ctx, task.id_var))

        return mk_primitive("rel_primitive_hash_tuple_uint128", [v for v, _ in terms])
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, (ir.Aggregate, ir.Output, ir.Update)):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    elif isinstance(task, ir.Rank):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

# Only used for translating terms on relatoms, which can be specialized values.
def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.Type]:
    if isinstance(term, ir.Literal) and term.type == types.Symbol:
        if isinstance(term.value, str):
            value = mk_value(term.value)
            return lqp.SpecializedValue(value=value, meta=None), meta_type_to_lqp(types.String)
        elif isinstance(term.value, int):
            value = mk_value(term.value)
            return lqp.SpecializedValue(value=value, meta=None), meta_type_to_lqp(types.Int64)
        else:
            raise NotImplementedError(f"Cannot specialize literal of type {type(term.value)}")
    return _translate_term(ctx, term)

def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.Type]:
    if isinstance(term, ir.Var):
        name = ctx.var_names.get_name_by_id(term.id, term.name)
        t = meta_type_to_lqp(term.type)
        return mk_var(name), t
    else:
        assert isinstance(term, ir.Literal), f"Cannot translate value {term!r} of type {type(term)} to LQP Term; neither Var nor Literal."
        v = to_lqp_value(term.value, term.type)
        return v, meta_type_to_lqp(term.type)

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    if task.relation == rel_builtins.cast:
        assert len(task.args) == 3, f"expected three terms for {task.relation.name}, got {len(task.args)}"

        terms = []
        for arg in task.args[1:]:
            term, _ = _translate_relterm(ctx, arg)
            terms.append(term)
        return lqp.Cast(input=terms[0], result=terms[1], meta=None)

    # TODO: want signature not name
    rel_name = task.relation.name
    terms = []
    term_types = []
    for arg in task.args:
        # Handle varargs, which come wrapped in a tuple.
        if isinstance(arg, tuple):
            for vararg in arg:
                term, ty = _translate_relterm(ctx, vararg)
                terms.append(term)
                term_types.append(ty)
        else:
            term, ty = _translate_relterm(ctx, arg)
            terms.append(term)
            term_types.append(ty)

    if rel_builtins.is_builtin(task.relation):
        if task.relation.name == "construct_datetime" and len(terms) == 7:
            # construct_datetime does not provide a timezone or milliseconds so we
            # default to 0 milliseconds and UTC timezone.
            lqp_name = relname_to_lqp_name(task.relation.name)
            extended_terms = [
                *terms[:-1],
                mk_value(0),
                mk_value("UTC"),
                terms[-1],
            ]
            return mk_primitive(lqp_name, extended_terms)
        elif task.relation.name == "like_match":
            # In the like_match API of QB, the variable is the first argument and the pattern the second,
            #   but in Rel, the expected argument order is reversed: pattern first, then variable.
            lqp_name = relname_to_lqp_name(task.relation.name)
            return mk_primitive(lqp_name, terms[::-1])
        elif task.relation.name == "parse_decimal" and len(terms) == 2:
            # parse_decimal in the metamodel is a binary relation, (value:String,
            # var:DecimalType) but we need to expand the lookup to (bits(var.precision),
            # var.scale, value, var), so we special case it here.
            lqp_name = relname_to_lqp_name(task.relation.name)
            value, var = task.args
            assert isinstance(value, ir.Node)
            assert isinstance(var, ir.Var)
            typ = var.type
            assert isinstance(typ, ir.DecimalType)
            bit_value = mk_value(types.digits_to_bits(typ.precision))
            precision_value = mk_value(typ.scale)
            extended_terms = [
                lqp.SpecializedValue(value=bit_value, meta=None),
                lqp.SpecializedValue(value=precision_value, meta=None),
                terms[0],
                terms[1],
            ]
            return mk_primitive(lqp_name, extended_terms)
        elif task.relation.name == "date_year":
            return _cast_primitive(ctx, task.relation.name, terms, term_types)
        elif task.relation.name == "date_month":
            return _cast_primitive(ctx, task.relation.name, terms, term_types)
        elif task.relation.name == "date_day":
            return _cast_primitive(ctx, task.relation.name, terms, term_types)
        elif rel_builtins.is_pragma(task.relation):
            lqp_name = pragma_to_lqp_name(task.relation.name)
            return mk_pragma(lqp_name, terms)
        else:
            lqp_name = relname_to_lqp_name(task.relation.name)
            if is_monotype(task.relation.name):
                # Make sure that the input terms have the same types
                assert term_types.count(term_types[0]) == len(term_types), \
                    f"Expected all terms to have the same type for monotype operator " \
                    f"`{task.relation.name}` but got {term_types} for terms {terms}"

            return mk_primitive(lqp_name, terms)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(name=task.relation.name, terms=terms, meta=None)

    rid = get_relation_id(ctx, rel_name, task.relation.id)
    return lqp.Atom(name=rid, terms=terms, meta=None)

def _cast_primitive(
        ctx: TranslationCtx,
        name: str,
        terms: list[lqp.RelTerm],
        term_types: list[lqp.Type]
    ) -> lqp.Formula:
    expected = primitive_types.get(name, None)
    if expected is None:
        return mk_primitive(relname_to_lqp_name(name), terms)
    elif expected[-1] == term_types[-1].type_name:
        return mk_primitive(relname_to_lqp_name(name), terms)
    # TODO: check the rest of the terms
    conversion_var = gen_unique_var(ctx, f"result_{expected[-1]}")
    result = terms[-1]
    assert isinstance(result, lqp.Var), f"Expected last term to be a Var, got {type(result)}: {result}"
    terms = [*terms[:-1], conversion_var]

    # Create a cast operation to convert the body to the result type
    cast = lqp.Cast(input=conversion_var, result=result, meta=None)
    primitive = mk_primitive(relname_to_lqp_name(name), terms)
    body = mk_and([primitive, cast])
    return mk_exists([(conversion_var, mk_type(expected[-1]))], body)

def get_relation_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> lqp.RelationId:
    mid_str = str(metamodel_id)
    relation_id = lqp.RelationId(id=utils.lqp_hash(mid_str), meta=None)
    unique_name = ctx.def_names.get_name_by_id(metamodel_id, orig_name)
    ctx.rel_id_to_orig_name[relation_id] = unique_name
    return relation_id

def _translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.Type, Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp.Var)
        return var, typ, None
    elif isinstance(binding, ir.Literal):
        lqp_value = to_lqp_value(binding.value, binding.type)
        var, formula = constant_to_var(ctx, lqp_value)
        return var, meta_type_to_lqp(binding.type), formula
    else:
        raise Exception(f"Unsupported binding type: {type(binding)}")

def to_lqp_value(value: ir.PyValue, value_type: ir.Type) -> lqp.Value:
    typ = meta_type_to_lqp(value_type)

    # Ensure int values match the requested integer type.
    if typ.type_name == lqp.TypeName.INT and isinstance(value, int):
        assert lqp_types.INT_MIN <= value <= lqp_types.INT_MAX, f"{value} out of range for a 64-bit INT value"
        val = value
    elif typ.type_name == lqp.TypeName.INT128 and isinstance(value, int):
        assert lqp_types.INT128_MIN <= value <= lqp_types.INT128_MAX, f"{value} out of range for an INT128 value"
        val = lqp.Int128Value(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.UINT128 and isinstance(value, int):
        assert lqp_types.UINT128_MIN <= value <= lqp_types.UINT128_MAX, f"{value} out of range for an UINT128 value"
        val = lqp.UInt128Value(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.FLOAT and isinstance(value, float):
        val = value
    elif typ.type_name == lqp.TypeName.STRING and isinstance(value, str):
        val = value
    elif typ.type_name == lqp.TypeName.DECIMAL and isinstance(value, PyDecimal):
        precision = typ.parameters[0].value
        scale = typ.parameters[1].value
        assert isinstance(precision, int) and isinstance(scale, int)
        val = lqp.DecimalValue(precision=precision, scale=scale, value=value, meta=None)
    elif typ.type_name == lqp.TypeName.DATE and isinstance(value, date):
        val = lqp.DateValue(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.DATETIME and isinstance(value, datetime):
        val = lqp.DateTimeValue(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.BOOLEAN and isinstance(value, bool):
        val = lqp.BooleanValue(value=value, meta=None)
    else:
        raise Exception(f"Unsupported type for constant: {typ.type_name} with value {value} of type {type(value)}")

    return mk_value(val)

def constant_to_var(ctx: TranslationCtx, value: lqp.Value, name_hint: str = "cvar") -> Tuple[lqp.Var, lqp.Formula]:
    var = gen_unique_var(ctx, name_hint)
    eq = mk_primitive("rel_primitive_eq", [var, value])
    return var, eq

def _extract_pyrel_error_ids(ctx: TranslationCtx, model: ir.Model) -> list[Tuple[lqp.RelationId, str]]:
    effects = collect_by_type(ir.Update, model)
    pyrel_error_attrs = []
    # We use a separate counter here to avoid modifying the current model.
    i = 1
    for effect in effects:
        if "pyrel_error_attrs" not in effect.relation.name:
            continue
        name = f"pyrel_error_attr_{i}"
        meta_id = effect.relation.id
        rel_id = get_relation_id(ctx, name, meta_id)
        pyrel_error_attrs.append((rel_id, name))
        i += 1

    return pyrel_error_attrs
